/*
 *   display.c
 *
 *   This file is part of Emu42
 *
 *   Copyright (C) 2004 Christoph Gieelink
 *
 */
#include "pch.h"
#include "resource.h"
#include "Emu42.h"
#include "ops.h"
#include "io.h"
#include "kml.h"

#define DISPLAY_FREQ	16					// display update 1/frequency (1/64) in ms

#define B 0x00000000						// black
#define W 0x00FFFFFF						// white
#define I 0xFFFFFFFF						// ignore

#define LCD_ROW_H		140					// max. pixel per line (131 or 137)
#define LCD_ROW_V32		 32					// max. pixel per line 32
#define LCD_ROW_V16		 16					// max. pixel per line 16

BOOL    bFullScreen = FALSE;
BOOL    bVertical = FALSE;
UINT    nBackgroundX = 0;
UINT    nBackgroundY = 0;
UINT    nBackgroundW = 0;
UINT    nBackgroundH = 0;
UINT    nLcdX = 0;
UINT    nLcdY = 0;
UINT    nLcdZoom = 1;
LPBYTE	pbyLcd;
HDC     hLcdDC = NULL;
HDC     hMainDC = NULL;

static HBITMAP hLcdBitmap;
static HBITMAP hMainBitmap;

static INT  nLcdxZoom,nLcdyZoom;
static INT  nLcdWidth,nLcdHeight;

static HANDLE hCThreadLcd = NULL;
static HANDLE hEventLcd;

static DWORD dwKMLColor[64] =				// color table loaded by KML script
{
	W,B,B,B,B,B,B,B,B,B,B,B,B,B,B,B,
	B,B,B,B,B,B,B,B,B,B,B,B,B,B,B,B,
	I,I,I,I,I,I,I,I,I,I,I,I,I,I,I,I,
	I,I,I,I,I,I,I,I,I,I,I,I,I,I,I,I
};

static struct
{
	BITMAPINFOHEADER Lcd_bmih;
	RGBQUAD bmiColors[2];
} bmiLcd =
{
	{sizeof(BITMAPINFOHEADER),0/*x*/,0/*y*/,1,8,BI_RGB,0,0,0,ARRAYSIZEOF(bmiLcd.bmiColors),0}
};

static VOID (*UpdateDisplayMem)(VOID) = NULL;

static VOID UpdateDisplayClamshellH(VOID);
static VOID UpdateDisplayPioneerH(VOID);
static VOID UpdateDisplayClamshellV(VOID);
static VOID UpdateDisplayPioneerV(VOID);

VOID UpdateContrast(BYTE byContrast)
{
	bmiLcd.bmiColors[1] = *(RGBQUAD*)&dwKMLColor[byContrast];    // pixel on  color
	bmiLcd.bmiColors[0] = *(RGBQUAD*)&dwKMLColor[byContrast+32]; // pixel off color
	if ((Chipset.IORam[DSPCTL] & DON) == 0 || I == *(DWORD*)&bmiLcd.bmiColors[0])
		bmiLcd.bmiColors[0] = *(RGBQUAD*)&dwKMLColor[0];

	// update palette information
	EnterCriticalSection(&csGDILock);		// asynchronous display update!
	{
		_ASSERT(hLcdDC);
		DeleteObject(SelectObject(hLcdDC,CreateDIBSection(hLcdDC,
			(BITMAPINFO*)&bmiLcd,DIB_RGB_COLORS,&pbyLcd,NULL,0)));
		ZeroMemory(pbyLcd,bmiLcd.Lcd_bmih.biWidth * nLcdHeight * nLcdyZoom * nAppZoom);
	}
	LeaveCriticalSection(&csGDILock);
	UpdateAnnunciators();					// adjust annunciator color
	return;
}

VOID SetLcdColor(UINT nId, UINT nRed, UINT nGreen, UINT nBlue)
{
	dwKMLColor[nId&0x3F] = ((nRed&0xFF)<<16)|((nGreen&0xFF)<<8)|(nBlue&0xFF);
	return;
}

VOID GetSizeLcdBitmap(INT *pnX, INT *pnY)
{
	_ASSERT(hLcdDC != NULL);				// following variables are valid
	*pnX = nLcdWidth  * nLcdxZoom;
	*pnY = nLcdHeight * nLcdyZoom;
	return;
}

VOID CreateLcdBitmap(VOID)
{
	INT nLcdBmpWidth;

	nLcdxZoom = nLcdyZoom = nLcdZoom;		// preset x/y zoom variables

	// create LCD bitmap
	if (bVertical)
	{
		if (cCurrentRomType != 'O' && cCurrentRomType != 'Y')
		{
			// set display update routine
			UpdateDisplayMem = UpdateDisplayPioneerV;

			nLcdBmpWidth = LCD_ROW_V16;
			nLcdWidth	 = 16;
			nLcdHeight   = 131;

			switch (nLcdyZoom)
			{
				default:
				case 1 : nLcdxZoom = 2; break;
				case 2 : nLcdxZoom = 3; break;
				case 3 : nLcdxZoom = 4; break;
				case 4 : nLcdxZoom = 6; break;
			}
		}
		else
		{
			// set display update routine
			UpdateDisplayMem = UpdateDisplayClamshellV;

			nLcdBmpWidth = LCD_ROW_V32;
			nLcdWidth	 = 32;
			nLcdHeight   = 137;
		}
	}
	else
	{
		nLcdBmpWidth = LCD_ROW_H;

		if (cCurrentRomType != 'O' && cCurrentRomType != 'Y')
		{
			// set display update routine
			UpdateDisplayMem = UpdateDisplayPioneerH;

			nLcdWidth  = 131;
			nLcdHeight = 16;

			switch (nLcdxZoom)
			{
				default:
				case 1 : nLcdyZoom = 2; break;
				case 2 : nLcdyZoom = 3; break;
				case 3 : nLcdyZoom = 4; break;
				case 4 : nLcdyZoom = 6; break;
			}
		}
		else
		{
			// set display update routine
			UpdateDisplayMem = UpdateDisplayClamshellH;

			nLcdWidth  = 137;
			nLcdHeight = 32;
		}
	}

	bmiLcd.Lcd_bmih.biWidth = nLcdBmpWidth * nLcdxZoom * nAppZoom;
	bmiLcd.Lcd_bmih.biHeight = nLcdHeight * nLcdyZoom * nAppZoom;
	if (!bVertical) bmiLcd.Lcd_bmih.biHeight = -bmiLcd.Lcd_bmih.biHeight;
	_ASSERT(hLcdDC == NULL);
	VERIFY(hLcdDC = CreateCompatibleDC(hWindowDC));
	VERIFY(hLcdBitmap = CreateDIBSection(hLcdDC,(BITMAPINFO*)&bmiLcd,DIB_RGB_COLORS,&pbyLcd,NULL,0));
	hLcdBitmap = SelectObject(hLcdDC,hLcdBitmap);
	_ASSERT(hPalette != NULL);
	SelectPalette(hLcdDC,hPalette,FALSE);	// set palette for LCD DC
	RealizePalette(hLcdDC);					// realize palette
	UpdateContrast(Chipset.contrast);		// fill with background color at startup
	return;
}

VOID DestroyLcdBitmap(VOID)
{
	// set contrast palette to startup colors
	WORD i = 0;   dwKMLColor[i++] = W;
	while(i < 32) dwKMLColor[i++] = B;
	while(i < 64) dwKMLColor[i++] = I;

	UpdateDisplayMem = NULL;

	if (hLcdDC != NULL)
	{
		// destroy LCD bitmap
		DeleteObject(SelectObject(hLcdDC,hLcdBitmap));
		DeleteDC(hLcdDC);
		hLcdDC = NULL;
		hLcdBitmap = NULL;
	}
	return;
}

BOOL CreateMainBitmap(LPCTSTR szFilename)
{
	_ASSERT(hWindowDC != NULL);
	VERIFY(hMainDC = CreateCompatibleDC(hWindowDC));
	if (hMainDC == NULL) return FALSE;		// quit if failed
	hMainBitmap = LoadBitmapFile(szFilename);
	if (hMainBitmap == NULL)
	{
		DeleteDC(hMainDC);
		return FALSE;
	}
	hMainBitmap = SelectObject(hMainDC,hMainBitmap);
	_ASSERT(hPalette != NULL);
	VERIFY(SelectPalette(hMainDC,hPalette,FALSE));
	RealizePalette(hMainDC);
	return TRUE;
}

VOID DestroyMainBitmap(VOID)
{
	if (hMainDC != NULL)
	{
		// destroy Main bitmap
		DeleteObject(SelectObject(hMainDC,hMainBitmap));
		DeleteDC(hMainDC);
		hMainDC = NULL;
		hMainBitmap = NULL;
	}
	return;
}

//****************
//*
//* LCD functions
//*
//****************

static __inline VOID WritePixelQuadH(BYTE *p, BYTE a)
{
	if (nLcdxZoom == 4)
	{
		CONST DWORD dwPattern[] = { 0x00000000, 0x01010101 };

		_ASSERT((DWORD) p % sizeof(DWORD) == 0); // check alignment for ARM CPU

		*(DWORD *)&p[0*LCD_ROW_H]  = *(DWORD *)&p[4*LCD_ROW_H]  =
		*(DWORD *)&p[8*LCD_ROW_H]  = *(DWORD *)&p[12*LCD_ROW_H] =
		dwPattern[a&1];

		*(DWORD *)&p[16*LCD_ROW_H] = *(DWORD *)&p[20*LCD_ROW_H] =
		*(DWORD *)&p[24*LCD_ROW_H] = *(DWORD *)&p[28*LCD_ROW_H] =
		dwPattern[(a>>1)&1];

		*(DWORD *)&p[32*LCD_ROW_H] = *(DWORD *)&p[36*LCD_ROW_H] =
		*(DWORD *)&p[40*LCD_ROW_H] = *(DWORD *)&p[44*LCD_ROW_H] =
		dwPattern[(a>>2)&1];

		*(DWORD *)&p[48*LCD_ROW_H] = *(DWORD *)&p[52*LCD_ROW_H] =
		*(DWORD *)&p[56*LCD_ROW_H] = *(DWORD *)&p[60*LCD_ROW_H] =
		dwPattern[(a>>3)&1];
		return;
	}
	if (nLcdxZoom == 3)
	{
		p[0*LCD_ROW_H+0]  = p[0*LCD_ROW_H+1]  = p[0*LCD_ROW_H+2]  =
		p[3*LCD_ROW_H+0]  = p[3*LCD_ROW_H+1]  = p[3*LCD_ROW_H+2]  =
		p[6*LCD_ROW_H+0]  = p[6*LCD_ROW_H+1]  = p[6*LCD_ROW_H+2]  = a&1;

		p[9*LCD_ROW_H+0]  = p[9*LCD_ROW_H+1]  = p[9*LCD_ROW_H+2]  =
		p[12*LCD_ROW_H+0] = p[12*LCD_ROW_H+1] = p[12*LCD_ROW_H+2] =
		p[15*LCD_ROW_H+0] = p[15*LCD_ROW_H+1] = p[15*LCD_ROW_H+2] = (a>>1)&1;

		p[18*LCD_ROW_H+0] = p[18*LCD_ROW_H+1] = p[18*LCD_ROW_H+2] =
		p[21*LCD_ROW_H+0] = p[21*LCD_ROW_H+1] = p[21*LCD_ROW_H+2] =
		p[24*LCD_ROW_H+0] = p[24*LCD_ROW_H+1] = p[24*LCD_ROW_H+2] = (a>>2)&1;

		p[27*LCD_ROW_H+0] = p[27*LCD_ROW_H+1] = p[27*LCD_ROW_H+2] = 
		p[30*LCD_ROW_H+0] = p[30*LCD_ROW_H+1] = p[30*LCD_ROW_H+2] =
		p[33*LCD_ROW_H+0] = p[33*LCD_ROW_H+1] = p[33*LCD_ROW_H+2] = (a>>3)&1;
		return;
	}
	if (nLcdxZoom == 2)
	{
		CONST WORD wPattern[] = { 0x0000, 0x0101 };

		_ASSERT((DWORD) p % sizeof(WORD) == 0); // check alignment for ARM CPU

		*(WORD *)&p[0*LCD_ROW_H]  =
		*(WORD *)&p[2*LCD_ROW_H]  = wPattern[a&1];

		*(WORD *)&p[4*LCD_ROW_H]  =
		*(WORD *)&p[6*LCD_ROW_H]  = wPattern[(a>>1)&1];

		*(WORD *)&p[8*LCD_ROW_H]  =
		*(WORD *)&p[10*LCD_ROW_H] = wPattern[(a>>2)&1];

		*(WORD *)&p[12*LCD_ROW_H] =
		*(WORD *)&p[14*LCD_ROW_H] = wPattern[(a>>3)&1];
		return;
	}
	if (nLcdxZoom == 1)
	{
		p[0*LCD_ROW_H] = a&1;
		p[1*LCD_ROW_H] = (a>>1)&1;
		p[2*LCD_ROW_H] = (a>>2)&1;
		p[3*LCD_ROW_H] = (a>>3)&1;
		return;
	}
	return;
}

static __inline VOID WritePixelRectH(BYTE *p, BYTE a)
{
	if (nLcdxZoom == 4)
	{
		CONST DWORD dwPattern[] = { 0x00000000, 0x01010101 };

		_ASSERT((DWORD) p % sizeof(DWORD) == 0); // check alignment for ARM CPU

		*(DWORD *)&p[0*LCD_ROW_H]  = *(DWORD *)&p[4*LCD_ROW_H]  =
		*(DWORD *)&p[8*LCD_ROW_H]  = *(DWORD *)&p[12*LCD_ROW_H] =
		*(DWORD *)&p[16*LCD_ROW_H] = *(DWORD *)&p[20*LCD_ROW_H] =
		dwPattern[a&1];

		*(DWORD *)&p[24*LCD_ROW_H] = *(DWORD *)&p[28*LCD_ROW_H] =
		*(DWORD *)&p[32*LCD_ROW_H] = *(DWORD *)&p[36*LCD_ROW_H] =
		*(DWORD *)&p[40*LCD_ROW_H] = *(DWORD *)&p[44*LCD_ROW_H] =
		dwPattern[(a>>1)&1];

		*(DWORD *)&p[48*LCD_ROW_H] = *(DWORD *)&p[52*LCD_ROW_H] =
		*(DWORD *)&p[56*LCD_ROW_H] = *(DWORD *)&p[60*LCD_ROW_H] =
		*(DWORD *)&p[64*LCD_ROW_H] = *(DWORD *)&p[68*LCD_ROW_H] =
		dwPattern[(a>>2)&1];

		*(DWORD *)&p[72*LCD_ROW_H] = *(DWORD *)&p[76*LCD_ROW_H] =
		*(DWORD *)&p[80*LCD_ROW_H] = *(DWORD *)&p[84*LCD_ROW_H] =
		*(DWORD *)&p[88*LCD_ROW_H] = *(DWORD *)&p[92*LCD_ROW_H] =
		dwPattern[(a>>3)&1];
		return;
	}
	if (nLcdxZoom == 3)
	{
		p[0*LCD_ROW_H+0]  = p[0*LCD_ROW_H+1]  = p[0*LCD_ROW_H+2]  =
		p[3*LCD_ROW_H+0]  = p[3*LCD_ROW_H+1]  = p[3*LCD_ROW_H+2]  =
		p[6*LCD_ROW_H+0]  = p[6*LCD_ROW_H+1]  = p[6*LCD_ROW_H+2]  =
		p[9*LCD_ROW_H+0]  = p[9*LCD_ROW_H+1]  = p[9*LCD_ROW_H+2]  = a&1;

		p[12*LCD_ROW_H+0] = p[12*LCD_ROW_H+1] = p[12*LCD_ROW_H+2] =
		p[15*LCD_ROW_H+0] = p[15*LCD_ROW_H+1] = p[15*LCD_ROW_H+2] =
		p[18*LCD_ROW_H+0] = p[18*LCD_ROW_H+1] = p[18*LCD_ROW_H+2] =
		p[21*LCD_ROW_H+0] = p[21*LCD_ROW_H+1] = p[21*LCD_ROW_H+2] = (a>>1)&1;

		p[24*LCD_ROW_H+0] = p[24*LCD_ROW_H+1] = p[24*LCD_ROW_H+2] =
		p[27*LCD_ROW_H+0] = p[27*LCD_ROW_H+1] = p[27*LCD_ROW_H+2] =
		p[30*LCD_ROW_H+0] = p[30*LCD_ROW_H+1] = p[30*LCD_ROW_H+2] =
		p[33*LCD_ROW_H+0] = p[33*LCD_ROW_H+1] = p[33*LCD_ROW_H+2] = (a>>2)&1;

		p[36*LCD_ROW_H+0] = p[36*LCD_ROW_H+1] = p[36*LCD_ROW_H+2] =
		p[39*LCD_ROW_H+0] = p[39*LCD_ROW_H+1] = p[39*LCD_ROW_H+2] =
		p[42*LCD_ROW_H+0] = p[42*LCD_ROW_H+1] = p[42*LCD_ROW_H+2] =
		p[45*LCD_ROW_H+0] = p[45*LCD_ROW_H+1] = p[45*LCD_ROW_H+2] = (a>>3)&1;
		return;
	}
	if (nLcdxZoom == 2)
	{
		CONST WORD wPattern[] = { 0x0000, 0x0101 };

		_ASSERT((DWORD) p % sizeof(WORD) == 0); // check alignment for ARM CPU

		*(WORD *)&p[0*LCD_ROW_H]  =
		*(WORD *)&p[2*LCD_ROW_H]  =
		*(WORD *)&p[4*LCD_ROW_H]  = wPattern[a&1];

		*(WORD *)&p[6*LCD_ROW_H]  =
		*(WORD *)&p[8*LCD_ROW_H]  =
		*(WORD *)&p[10*LCD_ROW_H] = wPattern[(a>>1)&1];

		*(WORD *)&p[12*LCD_ROW_H] =
		*(WORD *)&p[14*LCD_ROW_H] =
		*(WORD *)&p[16*LCD_ROW_H] = wPattern[(a>>2)&1];

		*(WORD *)&p[18*LCD_ROW_H] =
		*(WORD *)&p[20*LCD_ROW_H] =
		*(WORD *)&p[22*LCD_ROW_H] = wPattern[(a>>3)&1];
		return;
	}
	if (nLcdxZoom == 1)
	{
		p[0*LCD_ROW_H] = p[1*LCD_ROW_H] = a&1;
		p[2*LCD_ROW_H] = p[3*LCD_ROW_H] = (a>>1)&1;
		p[4*LCD_ROW_H] = p[5*LCD_ROW_H] = (a>>2)&1;
		p[6*LCD_ROW_H] = p[7*LCD_ROW_H] = (a>>3)&1;
		return;
	}
	return;
}

static VOID UpdateDisplayClamshellH(VOID)
{
	DWORD d;
	INT   x0,y0;
	BYTE  *p;

	// illegal zoom factor
	_ASSERT(nLcdxZoom == 1 || nLcdxZoom == 2 || nLcdxZoom == 3 || nLcdxZoom == 4);

	// scan complete display area of master
	for (d = MDISP_BEGIN; d <= MDISP_END; ++d)
	{
		// calculate x/y coordinates
		y0 = (d&0x7)*4;
		x0 = (SDISP_END-SDISP_BEGIN+1+d)>>3;

		// calculate memory position in LCD bitmap
		p = pbyLcd + (y0*LCD_ROW_H*nLcdyZoom + x0)*nLcdxZoom;

		// write pixel zoom independent
		WritePixelQuadH(p,Chipset.IORam[d]);
	}

	// scan complete display area of slave
	for (d = SDISP_BEGIN; d <= SDISP_END; ++d)
	{
		// calculate x/y coordinates
		y0 = ((d-SDISP_BEGIN)&0x7)*4;
		x0 = (d-SDISP_BEGIN)>>3;

		// calculate memory position in LCD bitmap
		p = pbyLcd + (y0*LCD_ROW_H*nLcdyZoom + x0)*nLcdxZoom;

		// write pixel zoom independent
		WritePixelQuadH(p,(BYTE)(ChipsetS.bLcdSyncErr ? 0xF : ChipsetS.IORam[d]));
	}
	return;
}

static VOID UpdateDisplayPioneerH(VOID)
{
	DWORD d;
	INT   x0,y0;
	BYTE  *p;

	// illegal zoom factor
	_ASSERT(nLcdxZoom == 1 || nLcdxZoom == 2 || nLcdxZoom == 3 || nLcdxZoom == 4);

	// scan complete display area
	for (d = DISP_BEGIN; d <= DISP_END; ++d)
	{
		// calculate x/y coordinates
		y0 = (d&0x3)*4;
		x0 = 66*((d>>2)&1)+(d>>3);

		// calculate memory position in LCD bitmap
		p = pbyLcd + (y0*LCD_ROW_H*nLcdyZoom + x0)*nLcdxZoom;

		// write pixel zoom independent
		WritePixelRectH(p,Chipset.IORam[d]);
	}
	return;
}

static __inline VOID WritePixelQuadV(BYTE *p, BYTE a)
{
	if (nLcdxZoom == 4)
	{
		CONST DWORD dwPattern[] = { 0x00000000, 0x01010101 };

		_ASSERT((DWORD) p % sizeof(DWORD) == 0); // check alignment for ARM CPU

		*(DWORD *)&p[ 0*LCD_ROW_V32+0] = *(DWORD *)&p[ 4*LCD_ROW_V32+0] =
		*(DWORD *)&p[ 8*LCD_ROW_V32+0] = *(DWORD *)&p[12*LCD_ROW_V32+0] =
		dwPattern[(a>>0)&1];

		*(DWORD *)&p[ 0*LCD_ROW_V32+4] = *(DWORD *)&p[ 4*LCD_ROW_V32+4] =
		*(DWORD *)&p[ 8*LCD_ROW_V32+4] = *(DWORD *)&p[12*LCD_ROW_V32+4] =
		dwPattern[(a>>1)&1];

		*(DWORD *)&p[ 0*LCD_ROW_V32+8] = *(DWORD *)&p[ 4*LCD_ROW_V32+8] =
		*(DWORD *)&p[ 8*LCD_ROW_V32+8] = *(DWORD *)&p[12*LCD_ROW_V32+8] =
		dwPattern[(a>>2)&1];

		*(DWORD *)&p[ 0*LCD_ROW_V32+12] = *(DWORD *)&p[ 4*LCD_ROW_V32+12] =
		*(DWORD *)&p[ 8*LCD_ROW_V32+12] = *(DWORD *)&p[12*LCD_ROW_V32+12] =
		dwPattern[(a>>3)&1];
		return;
	}
	if (nLcdxZoom == 3)
	{
		p[0*LCD_ROW_V32+0] = p[0*LCD_ROW_V32+1] = p[0*LCD_ROW_V32+2] =
		p[3*LCD_ROW_V32+0] = p[3*LCD_ROW_V32+1] = p[3*LCD_ROW_V32+2] =
		p[6*LCD_ROW_V32+0] = p[6*LCD_ROW_V32+1] = p[6*LCD_ROW_V32+2] =
		(a>>0)&1;

		p[0*LCD_ROW_V32+3] = p[0*LCD_ROW_V32+4] = p[0*LCD_ROW_V32+5] =
		p[3*LCD_ROW_V32+3] = p[3*LCD_ROW_V32+4] = p[3*LCD_ROW_V32+5] =
		p[6*LCD_ROW_V32+3] = p[6*LCD_ROW_V32+4] = p[6*LCD_ROW_V32+5] =
		(a>>1)&1;

		p[0*LCD_ROW_V32+6] = p[0*LCD_ROW_V32+7] = p[0*LCD_ROW_V32+8] =
		p[3*LCD_ROW_V32+6] = p[3*LCD_ROW_V32+7] = p[3*LCD_ROW_V32+8] =
		p[6*LCD_ROW_V32+6] = p[6*LCD_ROW_V32+7] = p[6*LCD_ROW_V32+8] =
		(a>>2)&1;

		p[0*LCD_ROW_V32+9] = p[0*LCD_ROW_V32+10] = p[0*LCD_ROW_V32+11] =
		p[3*LCD_ROW_V32+9] = p[3*LCD_ROW_V32+10] = p[3*LCD_ROW_V32+11] =
		p[6*LCD_ROW_V32+9] = p[6*LCD_ROW_V32+10] = p[6*LCD_ROW_V32+11] =
		(a>>3)&1;
		return;
	}
	if (nLcdxZoom == 2)
	{
		CONST DWORD dwPattern[] = { 0x00000000, 0x00000101, 0x01010000, 0x01010101 };

		_ASSERT((DWORD) p % sizeof(DWORD) == 0); // check alignment for ARM CPU

		*(DWORD *)&p[0*LCD_ROW_V32+0] =
		*(DWORD *)&p[2*LCD_ROW_V32+0] = dwPattern[(a>>0)&3];

		*(DWORD *)&p[0*LCD_ROW_V32+4] =
		*(DWORD *)&p[2*LCD_ROW_V32+4] = dwPattern[(a>>2)&3];
		return;
	}
	if (nLcdxZoom == 1)
	{
		p[0] = (a>>0)&1;
		p[1] = (a>>1)&1;
		p[2] = (a>>2)&1;
		p[3] = (a>>3)&1;
		return;
	}
	return;
}

static __inline VOID WritePixelRectV(BYTE *p, BYTE a)
{
	if (nLcdyZoom == 4)
	{
		CONST WORD  wPattern[]  = { 0x0000, 0x0101 };
		CONST DWORD dwPattern[] = { 0x00000000, 0x01010101 };

		_ASSERT((DWORD) p % sizeof(DWORD) == 0); // check alignment for ARM CPU

		*(DWORD *)&p[ 0*LCD_ROW_V16+0] = *(DWORD *)&p[ 6*LCD_ROW_V16+0] =
		*(DWORD *)&p[12*LCD_ROW_V16+0] = *(DWORD *)&p[18*LCD_ROW_V16+0] =
		dwPattern[(a>>0)&1];
		*(WORD *)&p[ 0*LCD_ROW_V16+4] = *(WORD *)&p[ 6*LCD_ROW_V16+4] =
		*(WORD *)&p[12*LCD_ROW_V16+4] = *(WORD *)&p[18*LCD_ROW_V16+4] =
		wPattern[(a>>0)&1];

		*(WORD *)&p[ 0*LCD_ROW_V16+6] = *(WORD *)&p[ 6*LCD_ROW_V16+6] =
		*(WORD *)&p[12*LCD_ROW_V16+6] = *(WORD *)&p[18*LCD_ROW_V16+6] =
		wPattern[(a>>1)&1];
		*(DWORD *)&p[ 0*LCD_ROW_V16+8] = *(DWORD *)&p[ 6*LCD_ROW_V16+8] =
		*(DWORD *)&p[12*LCD_ROW_V16+8] = *(DWORD *)&p[18*LCD_ROW_V16+8] =
		dwPattern[(a>>1)&1];

		*(DWORD *)&p[ 0*LCD_ROW_V16+12] = *(DWORD *)&p[ 6*LCD_ROW_V16+12] =
		*(DWORD *)&p[12*LCD_ROW_V16+12] = *(DWORD *)&p[18*LCD_ROW_V16+12] =
		dwPattern[(a>>2)&1];
		*(WORD *)&p[ 0*LCD_ROW_V16+16] = *(WORD *)&p[ 6*LCD_ROW_V16+16] =
		*(WORD *)&p[12*LCD_ROW_V16+16] = *(WORD *)&p[18*LCD_ROW_V16+16] =
		wPattern[(a>>2)&1];

		*(WORD *)&p[ 0*LCD_ROW_V16+18] = *(WORD *)&p[ 6*LCD_ROW_V16+18] =
		*(WORD *)&p[12*LCD_ROW_V16+18] = *(WORD *)&p[18*LCD_ROW_V16+18] =
		wPattern[(a>>3)&1];
		*(DWORD *)&p[ 0*LCD_ROW_V16+20] = *(DWORD *)&p[ 6*LCD_ROW_V16+20] =
		*(DWORD *)&p[12*LCD_ROW_V16+20] = *(DWORD *)&p[18*LCD_ROW_V16+20] =
		dwPattern[(a>>3)&1];
		return;
	}
	if (nLcdyZoom == 3)
	{
		p[0*LCD_ROW_V16+0] = p[0*LCD_ROW_V16+1] =
		p[0*LCD_ROW_V16+2] = p[0*LCD_ROW_V16+3] =
		p[4*LCD_ROW_V16+0] = p[4*LCD_ROW_V16+1] =
		p[4*LCD_ROW_V16+2] = p[4*LCD_ROW_V16+3] =
		p[8*LCD_ROW_V16+0] = p[8*LCD_ROW_V16+1] =
		p[8*LCD_ROW_V16+2] = p[8*LCD_ROW_V16+3] = (a>>0)&1;

		p[0*LCD_ROW_V16+4] = p[0*LCD_ROW_V16+5] =
		p[0*LCD_ROW_V16+6] = p[0*LCD_ROW_V16+7] =
		p[4*LCD_ROW_V16+4] = p[4*LCD_ROW_V16+5] =
		p[4*LCD_ROW_V16+6] = p[4*LCD_ROW_V16+7] =
		p[8*LCD_ROW_V16+4] = p[8*LCD_ROW_V16+5] =
		p[8*LCD_ROW_V16+6] = p[8*LCD_ROW_V16+7] = (a>>1)&1;

		p[0*LCD_ROW_V16+8]  = p[0*LCD_ROW_V16+9]  =
		p[0*LCD_ROW_V16+10] = p[0*LCD_ROW_V16+11] =
		p[4*LCD_ROW_V16+8]  = p[4*LCD_ROW_V16+9]  =
		p[4*LCD_ROW_V16+10] = p[4*LCD_ROW_V16+11] =
		p[8*LCD_ROW_V16+8]  = p[8*LCD_ROW_V16+9]  =
		p[8*LCD_ROW_V16+10] = p[8*LCD_ROW_V16+11] = (a>>2)&1;

		p[0*LCD_ROW_V16+12] = p[0*LCD_ROW_V16+13] =
		p[0*LCD_ROW_V16+14] = p[0*LCD_ROW_V16+15] =
		p[4*LCD_ROW_V16+12] = p[4*LCD_ROW_V16+13] =
		p[4*LCD_ROW_V16+14] = p[4*LCD_ROW_V16+15] =
		p[8*LCD_ROW_V16+12] = p[8*LCD_ROW_V16+13] =
		p[8*LCD_ROW_V16+14] = p[8*LCD_ROW_V16+15] = (a>>3)&1;
		return;
	}
	if (nLcdyZoom == 2)
	{
		p[0*LCD_ROW_V16+0] = p[0*LCD_ROW_V16+1]  = p[0*LCD_ROW_V16+2]  =
		p[3*LCD_ROW_V16+0] = p[3*LCD_ROW_V16+1]  = p[3*LCD_ROW_V16+2]  = (a>>0)&1;

		p[0*LCD_ROW_V16+3] = p[0*LCD_ROW_V16+4]  = p[0*LCD_ROW_V16+5]  =
		p[3*LCD_ROW_V16+3] = p[3*LCD_ROW_V16+4]  = p[3*LCD_ROW_V16+5]  = (a>>1)&1;

		p[0*LCD_ROW_V16+6] = p[0*LCD_ROW_V16+7]  = p[0*LCD_ROW_V16+8]  =
		p[3*LCD_ROW_V16+6] = p[3*LCD_ROW_V16+7]  = p[3*LCD_ROW_V16+8]  = (a>>2)&1;

		p[0*LCD_ROW_V16+9] = p[0*LCD_ROW_V16+10] = p[0*LCD_ROW_V16+11] =
		p[3*LCD_ROW_V16+9] = p[3*LCD_ROW_V16+10] = p[3*LCD_ROW_V16+11] = (a>>3)&1;
		return;
	}
	if (nLcdyZoom == 1)
	{
		CONST DWORD dwPattern[] = { 0x00000000, 0x00000101, 0x01010000, 0x01010101 };

		_ASSERT((DWORD) p % sizeof(DWORD) == 0); // check alignment for ARM CPU

		*(DWORD *)&p[0] = dwPattern[(a>>0)&3];
		*(DWORD *)&p[4] = dwPattern[(a>>2)&3];
		return;
	}
	return;
}

static VOID UpdateDisplayClamshellV(VOID)
{
	DWORD d;
	INT   x0,y0;
	BYTE  *p;

	// illegal zoom factor
	_ASSERT(nLcdyZoom == 1 || nLcdyZoom == 2 || nLcdyZoom == 4);

	// scan complete display area of master
	for (d = MDISP_BEGIN; d <= MDISP_END; ++d)
	{
		// calculate x/y coordinates
		y0 = (d&0x7)*4;
		x0 = (SDISP_END-SDISP_BEGIN+1+d)>>3;

		// calculate memory position in LCD bitmap
		p = pbyLcd + (x0*LCD_ROW_V32*nLcdyZoom + y0)*nLcdxZoom;

		// write pixel zoom independent
		WritePixelQuadV(p,Chipset.IORam[d]);
	}

	// scan complete display area of slave
	for (d = SDISP_BEGIN; d <= SDISP_END; ++d)
	{
		// calculate x/y coordinates
		y0 = ((d-SDISP_BEGIN)&0x7)*4;
		x0 = (d-SDISP_BEGIN)>>3;

		// calculate memory position in LCD bitmap
		p = pbyLcd + (x0*LCD_ROW_V32*nLcdyZoom + y0)*nLcdxZoom;

		// write pixel zoom independent
		WritePixelQuadV(p,(BYTE)(ChipsetS.bLcdSyncErr ? 0xF : ChipsetS.IORam[d]));
	}
	return;
}

static VOID UpdateDisplayPioneerV(VOID)
{
	DWORD d;
	INT   x0,y0;
	BYTE  *p;

	// illegal zoom factor
	_ASSERT(nLcdyZoom == 1 || nLcdyZoom == 2 || nLcdyZoom == 4);

	// scan complete display area
	for (d = DISP_BEGIN; d <= DISP_END; ++d)
	{
		// calculate x/y coordinates
		y0 = (d&0x3)*4;
		x0 = 66*((d>>2)&1)+(d>>3);

		// calculate memory position in LCD bitmap
		p = pbyLcd + (x0*LCD_ROW_V16*nLcdyZoom + y0)*nLcdxZoom;

		// write pixel zoom independent
		WritePixelRectV(p,Chipset.IORam[d]);
	}
	return;
}

static __inline VOID DoubleBitmap(VOID)
{
	CONST DWORD dwPattern[] = { 0x00000000, 0x00000101, 0x01010000, 0x01010101};

	INT x,y;

	// scan each row
	for (y = (nLcdHeight * nLcdyZoom - 1) * 2; y >= 0; y -= 2)
	{
		INT nOffset = y * bmiLcd.Lcd_bmih.biWidth;
		LPDWORD lpdwTarget = (LPDWORD) (pbyLcd + nOffset);
		LPBYTE  lpbySource = pbyLcd + nOffset / 4;

		// target address on DWORD boundary
		_ASSERT(((DWORD) lpdwTarget % sizeof(*lpdwTarget)) == 0);

		// scan each column
		for (x = bmiLcd.Lcd_bmih.biWidth / sizeof(*dwPattern); --x >= 0;)
		{
			// double two columns
			lpdwTarget[x] = dwPattern[lpbySource[x*2+1] * 2 + lpbySource[x*2]];
		}

		// double the line
		memcpy((LPBYTE)lpdwTarget+bmiLcd.Lcd_bmih.biWidth,lpdwTarget,bmiLcd.Lcd_bmih.biWidth);
	}
	return;
}

static DWORD WINAPI LcdThread(LPVOID pParam)
{
	while (WaitForSingleObject(hEventLcd,DISPLAY_FREQ) == WAIT_TIMEOUT)
	{
		EnterCriticalSection(&csGDILock);		// solving NT GDI problems
		{
			_ASSERT(UpdateDisplayMem);			// function not initialized
			UpdateDisplayMem();					// update display memory
			if (nAppZoom == 2) DoubleBitmap();	// double target bitmap

			BitBlt(hWindowDC,
				   (nLcdX - nBackgroundX) * nAppZoom, (nLcdY - nBackgroundY) * nAppZoom,
				   nLcdWidth * nLcdxZoom * nAppZoom, nLcdHeight * nLcdyZoom * nAppZoom,
				   hLcdDC,
				   0, 0,
				   SRCCOPY);
		}
		LeaveCriticalSection(&csGDILock);
	}
	return 0;
	UNREFERENCED_PARAMETER(pParam);
}

VOID UpdateAnnunciators(VOID)
{
	const DWORD dwAnnAddrP[] = { LA_ALL, LA_UPDOWN, LA_SHIFT, LA_PRINTER, LA_BUSY, LA_BAT, LA_G, LA_RAD };
	const DWORD dwAnnAddrO[] = { SLA_ALL, SLA_HALT, SLA_SHIFT, SLA_ALPHA, SLA_BUSY, SLA_BAT, SLA_RAD, SLA_PRINTER };

	DWORD  const *dwAnnAddr;
	LPBYTE pbySource;

	INT   i,j,nCount;

	BOOL bAnnOn = (Chipset.IORam[DSPCTL] & DON) != 0;

	if (Chipset.bSlave)						// has slave controller
	{
		dwAnnAddr = dwAnnAddrO;
		pbySource = ChipsetS.IORam;			// slave controller

		bAnnOn = bAnnOn && (ChipsetS.IORam[DSPCTL] & DON) != 0;
	}
	else
	{
		dwAnnAddr = dwAnnAddrP;
		pbySource = Chipset.IORam;			// master controller
	}

	// check all annuncators
	for (i = 1; i < ARRAYSIZEOF(dwAnnAddrP); ++i)
	{
		nCount = 0;

		if (bAnnOn)							// annunciators on
		{
			DWORD dwAnn = Npack(pbySource+dwAnnAddr[i],8) ^ Npack(pbySource+dwAnnAddr[0],8);

			// count the number of set bits
			for (;dwAnn != 0; dwAnn >>= 1)
			{
				if ((dwAnn & 0x1) != 0) ++nCount;
			}
		}

		// contrast table entry of annunciator
		j = Chipset.contrast + nCount - 16;
		if (j < 0)  j = 0;
		if (j > 31) j = 31;

		DrawAnnunciator(i, nCount > 0, dwKMLColor[j]);
	}
	return;
}

VOID StartDisplay(VOID)
{
	if (hCThreadLcd)						// LCD update thread running
		return;								// -> quit

	if (Chipset.IORam[DSPCTL]&DON)			// display on?
	{
		// event to cancel Lcd refresh loop
		hEventLcd = CreateEvent(NULL,FALSE,FALSE,NULL);

		UpdateAnnunciators();				// switch on annunciators
		VERIFY(hCThreadLcd = CreateThread(NULL,0,&LcdThread,NULL,0,NULL));
	}
	return;
}

VOID StopDisplay(VOID)
{
	// dummy region, WM_PAINT redraw always complete display area
	RECT Rect = { nLcdX, nLcdY, nLcdX+1, nLcdY+1 };

	if (hCThreadLcd == NULL)				// thread stopped
		return;								// -> quit

	SetEvent(hEventLcd);					// leave Lcd update thread
	WaitForSingleObject(hCThreadLcd,INFINITE);
	CloseHandle(hCThreadLcd);
	hCThreadLcd = NULL;						// set flag display update stopped
	CloseHandle(hEventLcd);					// close Lcd event

	ZeroMemory(pbyLcd, bmiLcd.Lcd_bmih.biWidth * nLcdHeight * nLcdyZoom * nAppZoom);
	InvalidateRect(hWnd,&Rect,FALSE);
	return;
}

VOID ResizeWindow(VOID)
{
	RECT rc;

	if (hWnd == NULL) return;				// return if window closed

	GetWindowRect(hWnd, &rc);

	// change client position
	rc.top = (bFullScreen) ? 0 : lrcTop;
	SetWindowPos(hWnd, NULL,
		rc.left,
		rc.top,
		nBackgroundW * nAppZoom,
		nBackgroundH * nAppZoom,
		SWP_NOZORDER);

	GetClientRect(hWnd, &rc);
	if (rc.bottom < (LONG) (nBackgroundH * nAppZoom))
	{
		SetWindowPos(hWnd, NULL, 0, 0,
			nBackgroundW * nAppZoom,
			(2 * nBackgroundH - rc.bottom) * nAppZoom,
			SWP_NOMOVE | SWP_NOZORDER);
	}
	InvalidateRect(hWnd,NULL,TRUE);
	return;
}
